]> git.sesse.net Git - cubemap/blobdiff - server.h
Write our own HTTP client instead of using curl. Not finished yet (missing URL parsin...
[cubemap] / server.h
index 2cb078e3fc418bdd6b47665bba31635b3af6e4e2..ce9365d7a49cfb98d176d9a566358c2a5f98d937 100644 (file)
--- a/server.h
+++ b/server.h
@@ -15,6 +15,7 @@
 
 class ClientProto;
 class CubemapStateProto;
+class Stream;
 class StreamProto;
 
 struct Client {
@@ -22,13 +23,13 @@ struct Client {
        Client(int sock);
 
        // Serialization/deserialization.
-       Client(const ClientProto &serialized);
+       Client(const ClientProto &serialized, Stream *stream);
        ClientProto serialize() const;
 
        // The file descriptor associated with this socket.
        int sock;
 
-       enum State { READING_REQUEST, SENDING_HEADER, SENDING_DATA };
+       enum State { READING_REQUEST, SENDING_HEADER, SENDING_DATA, SENDING_ERROR };
        State state;
 
        // The HTTP request, as sent by the client. If we are in READING_REQUEST,
@@ -38,14 +39,16 @@ struct Client {
        // What stream we're connecting to; parsed from <request>.
        // Not relevant for READING_REQUEST.
        std::string stream_id;
+       Stream *stream;
 
        // The header we want to send. This is nominally a copy of Stream::header,
        // but since that might change on reconnects etc., we keep a local copy here.
-       // Only relevant for SENDING_HEADER; blank otherwise.
-       std::string header;
+       // Only relevant for SENDING_HEADER or SENDING_ERROR; blank otherwise.
+       std::string header_or_error;
 
-       // Number of bytes we've sent of the header. Only relevant for SENDING_HEADER.
-       size_t header_bytes_sent;
+       // Number of bytes we've sent of the header. Only relevant for SENDING_HEADER
+       // or SENDING_ERROR.
+       size_t header_or_error_bytes_sent;
 
        // Number of bytes we've sent of data. Only relevant for SENDING_DATA.
        size_t bytes_sent;
@@ -70,6 +73,23 @@ struct Stream {
        // How many bytes <data> contains. Can very well be larger than BACKLOG_SIZE,
        // since the buffer wraps.
        size_t data_size;
+       
+       // Clients that are in SENDING_DATA, but that we don't listen on,
+       // because we currently don't have any data for them.
+       // See put_client_to_sleep() and wake_up_all_clients().
+       std::vector<Client *> sleeping_clients;
+
+       // Clients that we recently got data for (when they were in
+       // <sleeping_clients>).
+       std::vector<Client *> to_process;
+
+       // Put client to sleep, since there is no more data for it; we will on
+       // longer listen on POLLOUT until we get more data. Also, it will be put
+       // in the list of clients to wake up when we do.
+       void put_client_to_sleep(Client *client);
+
+       // We have more data, so mark all clients that are sleeping as ready to go.
+       void wake_up_all_clients();
 
 private:
        Stream(const Stream& other);
@@ -85,21 +105,43 @@ public:
 
        // Stop the thread.
        void stop();
+       
+       void set_header(const std::string &stream_id, const std::string &header);
 
-       CubemapStateProto serialize() const;
+       // These will be deferred until the next time an iteration in do_work() happens,
+       // and the order between them are undefined.
+       // XXX: header should ideally be ordered with respect to data.
+       void add_client_deferred(int sock);
+       void add_data_deferred(const std::string &stream_id, const char *data, size_t bytes);
 
-       void add_client(int sock);
+       // These should not be called while running, since that would violate
+       // threading assumptions (ie., that epoll is only called from one thread
+       // at the same time).
+       CubemapStateProto serialize();
        void add_client_from_serialized(const ClientProto &client);
-
        void add_stream(const std::string &stream_id);
        void add_stream_from_serialized(const StreamProto &stream);
 
-       void set_header(const std::string &stream_id, const std::string &header);
-       void add_data(const std::string &stream_id, const char *data, size_t bytes);
-
 private:
        pthread_t worker_thread;
 
+       // Mutex protecting queued_data only. Note that if you want to hold both this
+       // and <mutex> below, you will need to take <mutex> before this one.
+       pthread_mutex_t queued_data_mutex;
+
+       // Deferred commands that should be run from the do_work() thread as soon as possible.
+       // We defer these for two reasons:
+       //
+       //  - We only want to fiddle with epoll from one thread at any given time,
+       //    and doing add_client() from the acceptor thread would violate that.
+       //  - We don't want the input thread(s) hanging on <mutex> when doing
+       //    add_data(), since they want to do add_data() rather often, and <mutex>
+       //    can be taken a lot of the time.
+       //      
+       // Protected by <queued_data_mutex>.
+       std::vector<int> queued_add_clients;
+       std::map<std::string, std::string> queued_data;
+
        // All variables below this line are protected by the mutex.
        pthread_mutex_t mutex;
 
@@ -116,39 +158,48 @@ private:
        int epoll_fd;
        epoll_event events[EPOLL_MAX_EVENTS];
 
-       // Clients that are in SENDING_DATA, but that we don't listen on,
-       // because we currently don't have any data for them.
-       // See put_client_to_sleep() and wake_up_all_clients().
-       std::vector<int> sleeping_clients;
-
        // Recover the this pointer, and call do_work().
        static void *do_work_thunk(void *arg);
 
        // The actual worker thread.
        void do_work();
 
+       // Process a client; read and write data as far as we can.
+       // After this call, one of these four is true:
+       //
+       //  1. The socket is closed, and the client deleted.
+       //  2. We are still waiting for more data from the client.
+       //  3. We've sent all the data we have to the client,
+       //     and put it in <sleeping_clients>.
+       //  4. The socket buffer is full (which means we still have
+       //     data outstanding).
+       //
+       // For #2, we listen for EPOLLIN events. For #3 and #4, we listen
+       // for EPOLLOUT in edge-triggered mode; it will never fire for #3,
+       // but it's cheaper than taking it in and out all the time.
        void process_client(Client *client);
 
        // Close a given client socket, and clean up after it.
        void close_client(Client *client);
 
-       // Parse the HTTP request.
-       void parse_request(Client *client);
+       // Parse the HTTP request. Returns a HTTP status code (200/400/404).
+       int parse_request(Client *client);
 
        // Construct the HTTP header, and set the client into
        // the SENDING_HEADER state.
        void construct_header(Client *client);
 
-       // Put client to sleep, since there is no more data for it; we will on
-       // longer listen on POLLOUT until we get more data. Also, it will be put
-       // in the list of clients to wake up when we do.
-       void put_client_to_sleep(Client *client);
-
-       // We have more data, so mark all clients that are sleeping as ready to go.
-       void wake_up_all_clients();
+       // Construct a generic error with the given line, and set the client into
+       // the SENDING_ERROR state.
+       void construct_error(Client *client, int error_code);
 
        // TODO: This function should probably die.
        Stream *find_stream(const std::string &stream_id);
+
+       void process_queued_data();
+
+       void add_client(int sock);
+       void add_data(const std::string &stream_id, const char *data, size_t bytes);
 };
 
 #endif  // !defined(_SERVER_H)